import inspect
import logging
from typing import Union

from wiederverwendbar.logger.logger_singleton import SubLogger


class LoggingContext:
    class WrappedHandle:
        def __init__(self, saved_handle_method: callable):
            self.saved_handle_method = saved_handle_method
            self.contexts = []

        def __call__(self, logger, *args, **kwargs) -> None:
            # get stack
            stack = inspect.stack()[3:]

            # get all LoggingContexts in stack
            logging_contexts = []
            for frame_info in stack:
                frame = frame_info.frame
                for var in frame.f_locals.values():
                    if isinstance(var, LoggingContext):
                        if var.exited:
                            continue
                        logging_contexts.append(var)

            # filter out LoggingContexts that are not in self.contexts
            logging_contexts = [logging_context for logging_context in logging_contexts if logging_context in self.contexts]

            # raise error if logging_contexts is empty <- this should never happen
            if len(logging_contexts) == 0:
                raise RuntimeError("logging_contexts is empty")

            # check if some LoggingContexts need update
            for logging_context in logging_contexts:
                if logging_context.need_update:
                    logging_context.update()

            # get all loggers from logging_contexts
            context_loggers = [logging_context.context_logger for logging_context in logging_contexts]

            # handle to all logging_contexts
            for context_logger in context_loggers:
                # get context_logger class
                context_logger_type = type(context_logger)

                # get handle method of context_logger_type
                handle_method = getattr(context_logger_type, "handle")

                self.handle(handle_method, context_logger, *args, **kwargs)

            # handle with self.logger
            if logging_contexts[-1].handle_origin_logger:
                self.handle(self.saved_handle_method, *args, **kwargs)

        @classmethod
        def handle(cls, handle_method: callable, *args, **kwargs) -> None:
            # get signature of logger.handle
            signature = inspect.signature(handle_method)

            # bind signature to logger.handle
            bound_signature = signature.bind(*args, **kwargs)

            # handle with bound signature
            handle_method(**bound_signature.arguments)

    class ContextLogger(logging.Logger):
        def __new__(cls, *args, **kwargs):
            # get logging_contexts
            # get stack
            stack = inspect.stack()

            # get all LoggingContexts in stack
            logging_contexts = []
            for frame_info in stack:
                frame = frame_info.frame
                for var in frame.f_locals.values():
                    if isinstance(var, LoggingContext):
                        if var.exited:
                            continue
                        logging_contexts.append(var)

            # get logger class before context
            logger_class_before_context = None
            for logging_context in logging_contexts:
                _logger_class_before_context = logging_context._logger_class_before_context
                if _logger_class_before_context != cls:
                    if logger_class_before_context is not None and logger_class_before_context != _logger_class_before_context:
                        raise RuntimeError("logger_class_before_context is not None")
                    logger_class_before_context = _logger_class_before_context
            if logger_class_before_context is None:
                raise RuntimeError("logger_class_before_context is None")

            # use logger_class_before_context to build logger type
            logger = super().__new__(logger_class_before_context)

            # call __init__ of logger
            logger.__init__(*args, **kwargs)

            # update all logging contexts with logger
            for logging_context in logging_contexts:
                logging_context.update_one(logger)

            # set context_logger marker
            if isinstance(logger, SubLogger):
                with logger.reconfigure():
                    setattr(logger, "created_context_logger", True)
            else:
                setattr(logger, "created_context_logger", True)

            return logger

    def __init__(self, context_logger: logging.Logger, handle_origin_logger: bool = True):
        self.context_logger = context_logger

        # set context_logger marker
        if isinstance(self.context_logger, SubLogger):
            with self.context_logger.reconfigure():
                setattr(self.context_logger, "context_logger", True)
        else:
            setattr(self.context_logger, "context_logger", True)

        self.handle_origin_logger = handle_origin_logger
        self._exited = False
        self._wrapped_loggers: Union[tuple, tuple[logging.Logger]] = ()

        # get all loggers except context logger to prevent that existing loggers are ContextLogger
        self._get_all_loggers()

        # get current logger class
        self._logger_class_before_context = logging.getLoggerClass()
        if self._logger_class_before_context != LoggingContext.ContextLogger:
            logging.setLoggerClass(LoggingContext.ContextLogger)

    def __enter__(self) -> "LoggingContext":
        self.update()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        # reset _log method for all loggers
        for logger in self.wrapped_loggers:
            # get logger's handle method
            wrapped_handle: LoggingContext.WrappedHandle = getattr(logger, "handle")

            # skip if not wrapped
            if not hasattr(wrapped_handle, "saved_handle_method"):
                continue

            # raise error if context not in contexts
            if self not in wrapped_handle.contexts:
                raise RuntimeError(f"{self} is not in contexts")

            # remove context from wrapped_handle
            wrapped_handle.contexts.remove(self)

            # restore logger's handle method
            if len(wrapped_handle.contexts) == 0:
                if isinstance(logger, SubLogger):
                    with logger.reconfigure():
                        setattr(logger, "handle", wrapped_handle.saved_handle_method)
                else:
                    setattr(logger, "handle", wrapped_handle.saved_handle_method)

            # delete created_context_logger
            if getattr(logger, "created_context_logger", False):
                if logger.name in logging.root.manager.loggerDict:
                    logging.root.manager.loggerDict.pop(logger.name)

        # restore logger class
        if self._logger_class_before_context != LoggingContext.ContextLogger:
            logging.setLoggerClass(self._logger_class_before_context)

        # reset attributes
        self._wrapped_loggers = ()
        self._exited = True

    def _get_all_loggers(self) -> list[logging.Logger]:
        all_loggers = []
        for name in logging.root.manager.loggerDict:
            logger = logging.getLogger(name)
            # skip if self
            if logger == self.context_logger:
                continue
            # skip if logger is a context logger
            if getattr(logger, "context_logger", False):
                continue
            all_loggers.append(logger)
        return all_loggers

    @property
    def exited(self) -> bool:
        return self._exited

    @property
    def wrapped_loggers(self) -> Union[tuple, tuple[logging.Logger]]:
        return self._wrapped_loggers

    @property
    def need_update(self) -> bool:
        return not self._exited and len(self._wrapped_loggers) != len(self._get_all_loggers())

    def update(self) -> None:
        if self._exited:
            raise RuntimeError(f"{self} is already exited")

        # get all loggers except context logger
        all_loggers = self._get_all_loggers()

        # set _log method for all loggers
        for logger in all_loggers:
            self.update_one(logger)

    def update_one(self, logger: logging.Logger):
        wrapped_loggers = list(self._wrapped_loggers)

        # get logger's handle method
        logger_handle = getattr(logger, "handle")

        # check if already wrapped
        if hasattr(logger_handle, "saved_handle_method"):
            wrapped_handle: LoggingContext.WrappedHandle = logger_handle
            logger_handle = None
        else:
            wrapped_handle: LoggingContext.WrappedHandle = LoggingContext.WrappedHandle(logger_handle)

        # append context to wrapped_handle
        if self in wrapped_handle.contexts:
            return

        wrapped_handle.contexts.append(self)
        wrapped_loggers.append(logger)

        # overwrite logger's handle method
        if logger_handle is not None:
            if isinstance(logger, SubLogger):
                with logger.reconfigure():
                    setattr(logger, "handle", type(logger_handle)(wrapped_handle, logger))
            else:
                setattr(logger, "handle", type(logger_handle)(wrapped_handle, logger))

        self._wrapped_loggers = tuple(wrapped_loggers)



from copy import deepcopy
from typing import Dict, Any

from starlette.requests import Request
from starlette_admin.contrib.mongoengine.view import ModelView as BaseModelView

from wiederverwendbar.starlette_admin.mongoengine.generic_embedded_document_field.field import GenericEmbeddedDocumentField, ListField


class GenericEmbeddedDocumentView(BaseModelView):
    async def convert_generic_embedded_document(self, request: Request, data: Dict[str, Any], obj: Any) -> None:
        data = deepcopy(data)
        for field in self.fields:
            if isinstance(field, GenericEmbeddedDocumentField):
                value = data.get(field.name)
                if value is None:
                    continue
                value = await field.convert_generic_embedded_document(request, value)
                setattr(obj, field.name, value)
            elif isinstance(field, ListField) and isinstance(field.field, GenericEmbeddedDocumentField):
                for index, item in enumerate(data.get(field.name, [])):
                    value = await field.field.convert_generic_embedded_document(request, item)
                    data[field.name][index] = value
                setattr(obj, field.name, data[field.name])

    async def before_create(
            self, request: Request, data: Dict[str, Any], obj: Any
    ) -> None:
        await self.convert_generic_embedded_document(request, data, obj)

    async def before_edit(
            self, request: Request, data: Dict[str, Any], obj: Any
    ) -> None:
        await self.convert_generic_embedded_document(request, data, obj)



import inspect
import logging
from abc import ABCMeta
from functools import wraps
from typing import Optional

from wiederverwendbar.functions.find_class_method import find_class_method

logger = logging.getLogger(__name__)


class WrappedClass(ABCMeta):
    def __new__(cls, name, bases, attrs, order: Optional[int] = None):
        logger.debug(f"Creating {cls.__name__} '{name}'")

        # iterate over all attributes
        for attr_name, attr in attrs.items():
            if getattr(attr, "__ba_wrapped__", False):
                # get before and after names
                before_names = getattr(attr, "__ba_before__")
                after_names = getattr(attr, "__ba_after__")

                # get base attr if include_base is True
                base_before_names = []
                base_after_names = []
                if getattr(attr, "__ba_include_base__", False):
                    base_attr = find_class_method(bases, attr_name)
                    if base_attr is not None:
                        if getattr(base_attr, "__ba_wrapped__", False):
                            base_before_names = getattr(base_attr, "__ba_before__")
                            base_after_names = getattr(base_attr, "__ba_after__")
            else:
                # get base attr
                base_attr = find_class_method(bases, attr_name)
                if base_attr is None:
                    continue
                if not getattr(base_attr, "__ba_wrapped__", False):
                    continue
                if not getattr(base_attr, "__ba_include_inherited__", True):
                    continue

                # get before and after names from base attr
                before_names = []
                after_names = []
                base_before_names = getattr(base_attr, "__ba_before__")
                base_after_names = getattr(base_attr, "__ba_after__")

            # get before and after methods
            before_methods = _get_methods([*base_before_names, *before_names], attrs, bases)
            after_methods = _get_methods([*base_after_names, *after_names], attrs, bases)

            # wrap method
            attrs[attr_name] = _wrap(before_methods, after_methods)(attr)

        return ABCMeta.__new__(cls, name, bases, attrs)


def _get_methods(method_names: list[str], attrs: dict[str, callable], bases: list[type] = None):
    if bases is None:
        bases = []
    methods = []
    for method_name in method_names:
        if method_name in attrs:
            method = attrs[method_name]
        else:
            method = find_class_method(bases, method_name)
        if method is None:
            raise ValueError(f"Method '{method_name}' not found.")
        methods.append(method)
    return methods


def _get_signatures(func: callable, before_methods: list[callable], after_methods: list[callable]) -> tuple[inspect.Signature, list[inspect.Signature], list[inspect.Signature]]:
    # get before method signatures
    before_method_signatures = []
    for before_method in before_methods:
        if not callable(before_method):
            raise ValueError("before_method must be a callable.")

        logger.debug(f"Gather before_method '{before_method.__name__}' signature.")
        before_method_signatures.append(inspect.signature(before_method))

    # get after method signatures
    after_method_signatures = []
    for after_method in after_methods:
        if not callable(after_method):
            raise ValueError("after_method must be a callable.")

        logger.debug(f"Gather after_method '{after_method.__name__}' signature.")
        after_method_signatures.append(inspect.signature(after_method))

    # get func signature
    func_signature = inspect.signature(func)

    return func_signature, before_method_signatures, after_method_signatures


def _wrapper(*args,
             func: callable,
             before_methods: list[callable],
             after_methods: list[callable],
             func_signature: inspect.Signature,
             before_method_signatures: list[inspect.Signature],
             after_method_signatures: list[inspect.Signature],
             **kwargs):
    # add before_result to kwargs
    if "__ba_before_result__" in kwargs:
        raise ValueError("__ba_before_result__ is a reserved keyword.")
    kwargs["__ba_before_result__"] = None
    if "__ba_result__" in kwargs:
        raise ValueError("__ba_result__ is a reserved keyword.")
    kwargs["__ba_result__"] = None
    if "__ba_after_result__" in kwargs:
        raise ValueError("__ba_after_result__ is a reserved keyword.")
    kwargs["__ba_after_result__"] = None

    def bind(s: inspect.Signature) -> dict[str, any]:
        if "__ba_before_result__" in kwargs:
            if "__ba_before_result__" not in s.parameters:
                del kwargs["__ba_before_result__"]
        if "__ba_result__" in kwargs:
            if "__ba_result__" not in s.parameters:
                del kwargs["__ba_result__"]
        if "__ba_after_result__" in kwargs:
            if "__ba_after_result__" not in s.parameters:
                del kwargs["__ba_after_result__"]

        return s.bind(*args, **kwargs).arguments

    # get __ba_use__ flag
    __ba_use__ = kwargs.pop("__ba_use__", True)
    if not __ba_use__:
        mapped_params = bind(func_signature)
        return func(**mapped_params)

    # get __ba_use_before__ flag
    __ba_use_before__ = kwargs.pop("__ba_use_before__", True)

    # get __ba_use_after__ flag
    __ba_use_after__ = kwargs.pop("__ba_use_after__", True)

    # call before_methods
    if __ba_use_before__:
        for bm in before_methods:
            # get before_method_signature
            bms = before_method_signatures[before_methods.index(bm)]

            # map args and kwargs to before_method_signature
            mapped_params = bind(bms)

            # call before_method
            logger.debug(f"Call before_method '{bm.__name__}'")
            kwargs["__ba_before_result__"] = bm(**mapped_params)

    # map args and kwargs to func_signature
    mapped_params = bind(func_signature)

    # call func
    logger.debug(f"Call function '{func.__name__}'")
    __ba_result__ = func(**mapped_params)
    kwargs["__ba_result__"] = __ba_result__

    # call after_methods
    if __ba_use_after__:
        for am in after_methods:
            # get after_method_signature
            ams = after_method_signatures[after_methods.index(am)]

            # map args and kwargs to after_method_signature
            mapped_params = bind(ams)

            # call after_method
            logger.debug(f"Call after_method '{am.__name__}'")
            kwargs["__ba_after_result__"] = am(**mapped_params)

    return __ba_result__


def _wrap(before_methods: list[callable], after_methods: list[callable]):
    def decorator(func):
        # get signatures
        func_signature, before_method_signatures, after_method_signatures = _get_signatures(func, before_methods, after_methods)

        return wraps(func)(lambda *args, **kwargs: _wrapper(*args,
                                                            func=func,
                                                            before_methods=before_methods,
                                                            after_methods=after_methods,
                                                            func_signature=func_signature,
                                                            before_method_signatures=before_method_signatures,
                                                            after_method_signatures=after_method_signatures,
                                                            **kwargs))

    return decorator


def wrap(before=None, after=None, include_base=True, include_inherited=True):
    if before is None and after is None:
        raise ValueError("before and after cannot be None at the same time.")

    # get before names
    before_names = []
    if callable(before):
        # convert to list
        before = [before]
    if type(before) is str:
        before = [before]
    if before is not None:
        for b in before:
            if callable(b):
                before_names.append(b.__name__)
            elif type(b) is str:
                before_names.append(b)
            else:
                raise ValueError("before must be a list of callables or strings.")

    # get after names
    after_names = []
    if callable(after):
        # convert to list
        after = [after]
    if type(after) is str:
        after = [after]
    if after is not None:
        for a in after:
            if callable(a):
                after_names.append(a.__name__)
            elif type(a) is str:
                after_names.append(a)
            else:
                raise ValueError("after must be a list of callables or strings.")

    def decorator(func):
        # add wrapped flag to function
        func.__ba_wrapped__ = True

        # add before and after names to function
        func.__ba_before__ = before_names
        func.__ba_after__ = after_names

        # add include_base flag to function
        func.__ba_include_base__ = include_base

        # add include_inherited flag to function
        func.__ba_include_inherited__ = include_inherited

        # check if function is a class method
        if "." not in func.__qualname__:
            @wraps(func)
            def _module_wrap(*args, **kwargs):
                # get attrs from calling function
                attrs = inspect.currentframe().f_back.f_locals

                before_methods = _get_methods(before_names, attrs)
                after_methods = _get_methods(after_names, attrs)

                # get signatures
                func_signature, before_method_signatures, after_method_signatures = _get_signatures(func, before_methods, after_methods)

                return _wrapper(*args,
                                func=func,
                                before_methods=before_methods,
                                after_methods=after_methods,
                                func_signature=func_signature,
                                before_method_signatures=before_method_signatures,
                                after_method_signatures=after_method_signatures,
                                **kwargs)

            return _module_wrap

        return func

    return decorator



from datetime import datetime
from typing import Optional

from wiederverwendbar.task_manger.trigger import Trigger, Interval, At


class Task:
    def __init__(self,
                 payload,
                 manager=None,
                 name: Optional[str] = None,
                 trigger: Optional[Trigger] = None,
                 time_measurement_before_run: bool = True,
                 auto_add: Optional[bool] = None,
                 *args,
                 **kwargs):
        # set task name
        if name is None:
            name = payload.__name__
        self.name = name

        self.manager = None

        # set task trigger
        if not isinstance(trigger, (Interval, At)):
            raise ValueError("Invalid trigger object.")
        self.trigger = trigger

        # set task payload
        if not callable(payload):
            raise ValueError("Payload must be callable.")
        self._payload = payload

        # set payload args and kwargs
        self.args = []
        self.kwargs = {}
        if not iter(args):
            raise ValueError("Args must be iterable.")
        for arg in args:
            self.args.append(arg)
        if not isinstance(kwargs, dict):
            raise ValueError("Kwargs must be dict.")
        self.kwargs = kwargs

        self.time_measurement_before_run = time_measurement_before_run

        self._last_run: Optional[datetime] = None
        self._next_run: Optional[datetime] = None

        # indicate if task is done
        self._done = False

        # auto add task to manager
        if auto_add is None:
            auto_add = True if manager else False

        if auto_add:
            if not manager:
                raise ValueError("Manager object is required.")
            manager.add_task(self)

    def __str__(self):
        return f"{self.manager.name}.{self.name}"

    def init(self, manager):
        self.manager = manager
        self.trigger.init(manager)
        self.set_next_run()

        # log task creation
        self.manager.logger.debug(f"{self}: Task created.")

    @property
    def last_run(self) -> Optional[datetime]:
        if self._last_run is None:
            return datetime.fromtimestamp(0)
        return self._last_run

    @property
    def next_run(self) -> Optional[datetime]:
        if self.is_done:
            return None
        return self._next_run

    @property
    def is_done(self) -> bool:
        return self._done

    def set_next_run(self):
        if isinstance(self.trigger, Interval):
            next_run = self.trigger.next(self.last_run)
        elif isinstance(self.trigger, At):
            next_run = self.trigger.next()
            if next_run == self._next_run:
                self.done()
        else:
            raise ValueError("Invalid trigger object.")
        self._next_run = next_run

    def set_last_run(self):
        self._last_run = datetime.now()

    def done(self):
        self._done = True

    def payload(self):
        self._payload(*self.args, **self.kwargs)



import logging
import sys

from wiederverwendbar.logger import LoggingContext
from wiederverwendbar.logger import LoggerSingleton, LoggerSettings

from wiederverwendbar.examples.logger_context.example_module import example_function

LoggerSingleton(init=True, name="test", settings=LoggerSettings(log_level="DEBUG"), ignored_loggers_like=["logger1", "logger2"])

logger1 = logging.getLogger("logger1")
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
ch.setFormatter(logging.Formatter('logger1 - %(name)s - %(levelname)s - %(message)s'))
logger1.addHandler(ch)
logger1.setLevel(logging.DEBUG)

logger2 = logging.getLogger("logger2")
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
ch.setFormatter(logging.Formatter('logger2 - %(name)s - %(levelname)s - %(message)s'))
logger2.addHandler(ch)
logger2.setLevel(logging.DEBUG)

if __name__ == '__main__':
    # no context
    logger1.debug("no context")
    logger2.debug("no context")

    with LoggingContext(logger1, handle_origin_logger=False) as context1_a:
        with LoggingContext(logger2, handle_origin_logger=False) as context2_a:
            example_function()
        example_function()
    example_function()

    with LoggingContext(logger1, handle_origin_logger=False) as context1_b:
        with LoggingContext(logger2, handle_origin_logger=False) as context2_b:
            example_function()
        example_function()
    example_function()

    # again no context
    logger1.debug("again no context")
    logger2.debug("again no context")

    sys.exit(0)



from wiederverwendbar.starlette_admin.action_log.logger import WebsocketHandler, ActionSubLogger, ActionSubLoggerContext, ActionLogger
from wiederverwendbar.starlette_admin.action_log.admin import ActionLogAdmin



from wiederverwendbar.starlette_admin.admin import MultiPathAdmin, FormMaxFieldsAdmin

from wiederverwendbar.starlette_admin.view import PropertyModelView
from wiederverwendbar.starlette_admin.view import property_field

from wiederverwendbar.starlette_admin.mongoengine import FixedModelView

from wiederverwendbar.starlette_admin.mongoengine import GenericEmbeddedAdmin
from wiederverwendbar.starlette_admin.mongoengine import GenericEmbeddedConverter
from wiederverwendbar.starlette_admin.mongoengine import GenericEmbeddedDocumentField, ListField
from wiederverwendbar.starlette_admin.mongoengine import GenericEmbeddedDocumentView

from wiederverwendbar.starlette_admin.mongoengine import IPv4Converter

from wiederverwendbar.starlette_admin.action_log import WebsocketHandler, ActionSubLogger, ActionSubLoggerContext, ActionLogger
from wiederverwendbar.starlette_admin.action_log import ActionLogAdmin



from rich.console import Console
from rich.logging import RichHandler

from wiederverwendbar.logger.handlers.stream_console_handler import _resolve_file
from wiederverwendbar.logger.logger_settings import LoggerSettings


class RichConsoleHandler(RichHandler):
    def __init__(self, *args, name: str, console_outfile: LoggerSettings.TerminalOutFile, console_width: int, **kwargs):
        super().__init__(
            *args,
            console=Console(file=_resolve_file(console_outfile),
                            width=console_width),
            **kwargs
        )
        self.set_name(name)


from wiederverwendbar.before_after_wrap import WrappedClass, wrap


class BaseClass(metaclass=WrappedClass):
    def before_cls_test1(self, __ba_before_result__):
        print("before_cls_test1")

    @wrap(before=before_cls_test1, after="after_cls_test1")
    def cls_test1(self, __ba_before_result__):
        print("cls_test1")

    def after_cls_test1(self, __ba_result__):
        print("after_cls_test1")

    def before_cls_test2(self):
        print("before_cls_test2")

    @wrap(before=before_cls_test2, after="after_cls_test2")
    def cls_test2(self):
        print("cls_test2")

    def after_cls_test2(self):
        print("after_cls_test2")


class ChildClass(BaseClass):
    def before_cls_test1A(self):
        print("before_cls_test1A")

    @wrap(before=before_cls_test1A, after="after_cls_test1A", include_base=False)
    def cls_test1(self, __ba_before_result__):
        super().cls_test1(__ba_use__=False)
        print("cls_test1A")

    def after_cls_test1A(self):
        print("after_cls_test1A")

    def before_cls_test3(self):
        print("before_cls_test3")

    @wrap(before=before_cls_test3, after="after_cls_test3")
    def cls_test3(self):
        print("cls_test3")

    def after_cls_test3(self):
        print("after_cls_test3")


def before_test1(*_, **__):
    print("before_test1")


@wrap(before=before_test1, after="after_test1")
def test1(arg1: str):
    print(f"test1 -> {arg1}")


def after_test1(*_, **__):
    print("after_test1")


if __name__ == "__main__":
    base_class = BaseClass()
    base_class.cls_test1()
    base_class.cls_test2()

    child_class = ChildClass()
    child_class.cls_test1()
    child_class.cls_test2()
    child_class.cls_test3()

    test1("qwe")





from dataclasses import dataclass
from dataclasses import field as dc_field
from typing import Any, Dict, Type, Sequence, List, Optional

from starlette.requests import Request
from starlette.datastructures import FormData
from starlette_admin import RequestAction
from starlette_admin.helpers import extract_fields

import mongoengine as me
import starlette_admin as sa


@dataclass
class GenericEmbeddedDocumentField(sa.BaseField):
    embedded_doc_name_mapping: Dict[str, Type[me.EmbeddedDocument]] = dc_field(default_factory=dict)
    embedded_doc_fields: Dict[str, Sequence[sa.BaseField]] = dc_field(default_factory=dict)
    render_function_key: str = "json"
    form_template: str = "forms/generic_embedded.html"
    display_template: str = "displays/generic_embedded.html"
    select2: bool = True

    def __post_init__(self) -> None:
        super().__post_init__()
        self._propagate_id()

    @property
    def fieldset_id(self) -> str:
        if self.id == "":
            return "fieldset"
        return self.id + ".fieldset"

    def get_fields_list(
            self,
            request: Request,
            doc_name: str,
            action: RequestAction = RequestAction.LIST,
    ) -> Sequence[sa.BaseField]:
        _fields = []
        for current_doc_name, fields in self.embedded_doc_fields.items():
            if doc_name != "" and doc_name != current_doc_name:
                continue
            _fields.extend(extract_fields(fields, action))
        return _fields

    def _propagate_id(self) -> None:
        """Will update fields id by adding his id as prefix (ex: category.name)"""
        for doc_name, doc_fields in self.embedded_doc_fields.items():
            for field in doc_fields:
                field.id = self.id + ("." if self.id else "") + doc_name + "." + field.name
                if isinstance(field, type(self)):
                    field._propagate_id()

    def _add_doc_data(self, doc_data: Dict[str, Any], doc_name: str, doc_type: Type[me.EmbeddedDocument]) -> Dict[str, Any]:
        # add doc name to serialized value at detail and edit view
        if "__doc_name__" in doc_data:
            raise ValueError(f"Field name '__doc_name__' is reserved")
        doc_data["__doc_name__"] = doc_name

        return doc_data

    async def convert_generic_embedded_document(self, request: Request, value: Dict[str, Any]) -> Any:
        doc_name = value.pop("__doc_name__", "")
        if doc_name == "":
            raise ValueError(f"Field name '__doc_name__' not found in value: {value}")
        doc_type = self.embedded_doc_name_mapping.get(doc_name, None)
        if doc_type is None:
            raise ValueError(f"Invalid choice value: {doc_name}")

        embedded_doc = doc_type(**value)
        return embedded_doc

    async def parse_form_data(
            self, request: Request, form_data: FormData, action: RequestAction
    ) -> Any:
        # get selection field value
        doc_name = form_data.get(self.id, "")
        if doc_name == "":
            return None
        doc_type = self.embedded_doc_name_mapping.get(doc_name, None)
        if doc_type is None:
            raise ValueError(f"Invalid choice value: {doc_name}")

        # get selected embedded document fields
        doc_data = {}
        for field in self.get_fields_list(request, doc_name, action):
            doc_data[field.name] = await field.parse_form_data(request, form_data, action)

        # add doc name to serialized value at detail and edit view
        if "__doc_name__" in doc_data:
            raise ValueError(f"Field name '__doc_name__' is reserved")
        if action == RequestAction.EDIT or action == RequestAction.DETAIL:
            doc_data["__doc_name__"] = doc_name

        return doc_data

    async def serialize_value(
            self, request: Request, value: Any, action: RequestAction
    ) -> Any:
        # get embedded document name
        doc_name = None
        doc_type = None
        for name, doc_type in self.embedded_doc_name_mapping.items():
            if isinstance(value, doc_type):
                doc_name = name
                doc_type = doc_type
                break
        if doc_name is None or doc_type is None:
            raise ValueError(f"Invalid embedded document value: {value}")

        # return doc name at list view if render function is text
        if action == RequestAction.LIST and self.render_function_key == "text":
            return doc_name

        serialized_value: Dict[str, Any] = {}
        for field in self.get_fields_list(request, doc_name, action):
            name = field.name
            serialized_value[name] = None
            if hasattr(value, name) or (isinstance(value, dict) and name in value):
                field_value = (
                    getattr(value, name) if hasattr(value, name) else value[name]
                )
                if field_value is not None:
                    serialized_value[name] = await field.serialize_value(
                        request, field_value, action
                    )

        # add doc name to serialized value at detail and edit view
        if "__doc_name__" in serialized_value:
            raise ValueError(f"Field name '__doc_name__' is reserved")
        if action == RequestAction.EDIT or action == RequestAction.DETAIL:
            serialized_value["__doc_name__"] = doc_name
        return serialized_value

    def additional_css_links(
            self, request: Request, action: RequestAction
    ) -> List[str]:
        _links = []
        if self.select2 and action.is_form():
            _links.append(
                str(
                    request.url_for(
                        f"{request.app.state.ROUTE_NAME}:statics",
                        path="css/select2.min.css",
                    )
                )
            )
        for f in self.get_fields_list(request, "", action):
            _links.extend(f.additional_css_links(request, action))
        return _links

    def additional_js_links(self, request: Request, action: RequestAction) -> List[str]:
        _links = []
        if self.select2 and action.is_form():
            _links.append(
                str(
                    request.url_for(
                        f"{request.app.state.ROUTE_NAME}:statics",
                        path="js/vendor/select2.min.js",
                    )
                )
            )
            _links.append(
                str(
                    request.url_for(
                        f"{request.app.state.ROUTE_NAME}:statics",
                        path="js/generic_embedded.js",
                    )
                )
            )
        for f in self.get_fields_list(request, "", action):
            _links.extend(f.additional_js_links(request, action))
        return _links


@dataclass(init=False)
class ListField(sa.ListField):
    def __post_init__(self) -> None:
        super().__post_init__()
        if isinstance(self.field, GenericEmbeddedDocumentField):
            self.field._propagate_id()

    async def parse_form_data(
            self, request: Request, form_data: FormData, action: RequestAction
    ) -> Any:
        indices = self._extra_indices(form_data)
        value = []
        for index in indices:
            self.field.id = f"{self.id}.{index}"
            if isinstance(self.field, sa.CollectionField):
                self.field._propagate_id()
            if isinstance(self.field, GenericEmbeddedDocumentField):
                self.field._propagate_id()
            value.append(await self.field.parse_form_data(request, form_data, action))
        return value

    def _field_at(self, idx: Optional[int] = None) -> sa.BaseField:
        super()._field_at(idx)
        if isinstance(self.field, GenericEmbeddedDocumentField):
            self.field._propagate_id()
        return self.field



from typing import Optional

from pydantic import BaseModel

from wiederverwendbar.pydantic.printable_settings import Field


class MongoengineSettings(BaseModel):
    db_host: str = Field(default="localhost",
                         title="Database Host",
                         description="Host to connect to database")
    db_port: int = Field(default=27017,
                         title="Database Port",
                         ge=0,
                         le=65535,
                         description="Port to connect to database")
    db_name: str = Field(default=...,
                         title="Database Name",
                         description="Name of the database")
    db_username: Optional[str] = Field(default=None,
                                       title="Database User",
                                       description="User to connect to database")
    db_password: Optional[str] = Field(None,
                                       title="Database Password",
                                       description="Password to connect to database",
                                       secret=True)
    db_timeout: int = Field(default=1000,
                            title="Database Timeout",
                            ge=0,
                            le=60000,
                            description="Timeout to connect to database in milliseconds")



from typing import Union

from jinja2 import BaseLoader, ChoiceLoader, FileSystemLoader, PackageLoader
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles
from starlette_admin.base import BaseAdmin


class MultiPathAdminMeta(type):
    def __new__(cls, name, bases, attrs):
        # combine static_files_packages from bases and attrs
        all_static_files_packages = []
        # get all static_files_packages from bases
        for base in bases:
            if hasattr(base, "static_files_packages"):
                all_static_files_packages.extend(base.static_files_packages)
        # get static_files_packages from attrs
        if "static_files_packages" in attrs:
            for static_files_package in attrs["static_files_packages"]:
                # skip duplicates
                if static_files_package in all_static_files_packages:
                    continue
                # add at the beginning of the combined list
                all_static_files_packages.insert(0, static_files_package)

        # set static_files_packages to the combined list
        attrs["static_files_packages"] = all_static_files_packages

        # combine template_packages from bases and attrs
        all_template_packages = []
        # get all template_packages from bases
        for base in bases:
            if hasattr(base, "template_packages"):
                all_template_packages.extend(base.template_packages)

        # get template_packages from attrs
        if "template_packages" in attrs:
            for template_package in attrs["template_packages"]:
                # skip duplicates
                if template_package in all_template_packages:
                    continue
                # add at the beginning of the combined list
                all_template_packages.insert(0, template_package)

        # set template_packages to the combined list
        attrs["template_packages"] = all_template_packages

        return super().__new__(cls, name, bases, attrs)


class MultiPathAdmin(BaseAdmin, metaclass=MultiPathAdminMeta):
    """
    A base class for Admin classes that can be used in multiple paths.
    In detail, it combines the static_files_packages and template_packages from all bases and attrs.
    """

    static_files_packages: list[Union[str, tuple[str, str]]] = ["starlette_admin"]
    template_packages: list[BaseLoader] = [PackageLoader("starlette_admin", "templates")]

    def init_routes(self) -> None:
        super().init_routes()

        # find the statics mount index
        statics_index = None
        for i, route in enumerate(self.routes):
            if isinstance(route, Mount) and route.name == "statics":
                statics_index = i
                break
        if statics_index is None:
            raise ValueError("Could not find statics mount")

        # override the static files route
        self.routes[statics_index] = Mount("/statics", app=StaticFiles(directory=self.statics_dir, packages=self.static_files_packages), name="statics")

    def _setup_templates(self) -> None:
        super()._setup_templates()

        self.templates.env.loader = ChoiceLoader(
            [
                FileSystemLoader(self.templates_dir),
                *self.template_packages,
            ]
        )


class FormMaxFieldsAdmin(BaseAdmin):
    form_max_fields: int = 1000

    async def _render_create(self, request: Request) -> Response:
        self._form_func = request.form
        request.form = self.form

        return await super()._render_create(request)

    async def _render_edit(self, request: Request) -> Response:
        self._form_func = request.form
        request.form = self.form

        return await super()._render_edit(request)

    async def form(self, *args, **kwargs):
        if "max_fields" not in kwargs:
            kwargs["max_fields"] = self.form_max_fields

        return await self._form_func(*args, **kwargs)



from wiederverwendbar.logger.logger import Logger
from wiederverwendbar.logger.logger_settings import LoggerSettings
from wiederverwendbar.logger.logger_singleton import LoggerSingleton
from wiederverwendbar.logger.logger_context import LoggingContext



from wiederverwendbar.starlette_admin.mongoengine.generic_embedded_document_field.admin import GenericEmbeddedAdmin
from wiederverwendbar.starlette_admin.mongoengine.generic_embedded_document_field.converter import GenericEmbeddedConverter
from wiederverwendbar.starlette_admin.mongoengine.generic_embedded_document_field.field import GenericEmbeddedDocumentField, ListField
from wiederverwendbar.starlette_admin.mongoengine.generic_embedded_document_field.view import GenericEmbeddedDocumentView



from typing import Callable, Type
from starlette_admin.fields import BaseField


def property_field(field: Type[BaseField], **kwargs):
    def decorator(fget: Callable):
        fget.__sa_field_type__ = field
        kwargs["name"] = fget.__name__
        fget.__sa_field_kwargs__ = kwargs
        p = property(fget)
        return p

    return decorator



from wiederverwendbar.logger import LoggerSingleton, LoggerSettings
from wiederverwendbar.singleton import Singleton

logger = LoggerSingleton(init=True, name="test", settings=LoggerSettings(log_level="DEBUG"), ignored_loggers_like=["uvicorn"])


class A(metaclass=Singleton, order=3):
    def __init__(self):
        print("A created")

    def __del__(self):
        print("A deleted")


class B(metaclass=Singleton, order=2):
    def __init__(self):
        print("B created")

    def __del__(self):
        print("B deleted")


class C(metaclass=Singleton, order=1):
    def __init__(self):
        print("C created")

    def __del__(self):
        print("C deleted")


def init():
    A(init=True)
    B(init=True)
    C(init=True)
    print("init")


def main():
    a = A()
    b = B()
    c = C()

    print("main")


if __name__ == '__main__':
    Singleton.delete_ordered_on_exit = True
    init()
    main()


import logging

logger = logging.getLogger(__name__)


def example_function():
    logger.debug("example_function")
    func_logger = logging.getLogger(__name__ + ".example_function")
    func_logger.debug("debug")
    return



from typing import Any, Union

from pydantic import BaseModel


class IndexableModel(BaseModel):
    def __getitem__(self, item: Union[int, str]) -> Any:
        """
        Get item from model

        :param item: Index or key
        :return: Index value or key value
        """

        if isinstance(item, int):
            # get field list
            fields = list(self.model_fields.keys())

            # get field name
            field_name = fields[item]

            # get field value
            value = self.__getattribute__(field_name)
        elif isinstance(item, str):
            # get field value
            value = self.__getattribute__(item)
        else:
            raise TypeError(f"Index must be int or str not {type(item)}")

        return value


from wiederverwendbar.pydantic.indexable_model import IndexableModel


class Config(IndexableModel):
    asd: int = 123
    qwe: str = "qwe"
    yxc: bool = False


if __name__ == '__main__':
    config = Config()

    asd = config["asd"]
    qwe = config["qwe"]
    yxc = config["yxc"]

    asd_1 = config[0]
    qwe_1 = config[1]
    yxc_1 = config[2]



import logging

from jinja2 import PackageLoader
from starlette.routing import WebSocketRoute
from starlette.endpoints import WebSocketEndpoint
from starlette.websockets import WebSocket

from wiederverwendbar.starlette_admin.admin import MultiPathAdmin
from wiederverwendbar.starlette_admin.action_log.logger import ActionLogger

logger = logging.getLogger(__name__)


class ActionLogAdmin(MultiPathAdmin):
    static_files_packages = [("wiederverwendbar", "starlette_admin/action_log/statics")]
    template_packages = [PackageLoader("wiederverwendbar", "starlette_admin/action_log/templates")]

    class ActionLogEndpoint(WebSocketEndpoint):
        encoding = "text"
        wait_for_logger_timeout = 5

        async def on_connect(self, websocket: WebSocket):
            await websocket.accept()

            # wait for action logger
            try:
                action_logger = await ActionLogger.wait_for_logger(websocket, timeout=self.wait_for_logger_timeout)
            except ValueError:
                await websocket.close(code=1008)
                return

            # add websocket to action logger
            action_logger.add_websocket(websocket)

        async def on_disconnect(self, websocket: WebSocket, close_code: int):
            # get logger
            action_logger = await ActionLogger.get_logger(websocket)
            if action_logger is None:
                return

            # remove websocket from logger
            action_logger.remove_websocket(websocket)

    def init_routes(self) -> None:
        super().init_routes()
        self.routes.append(WebSocketRoute(path="/ws/action_log/{action_log_key}", endpoint=self.ActionLogEndpoint, name="action_log"))  # noqa



import logging
import time

from wiederverwendbar.task_manger import Manager, Task, AtCreation, EverySeconds

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
logger.addHandler(ch)

# create manager
manager1 = Manager(name="Manager1", logger=logger)
manager2 = Manager(name="Manager2", logger=logger)


@manager1.task(name="Task 1", trigger=AtCreation(delay_for_seconds=10))
def task1():
    logger.debug("Task 1 ...")


@manager1.task(name="Task 2", trigger=EverySeconds(2))
def task2():
    logger.debug("Task 2 ...")
    time.sleep(2)


@manager1.task(name="Task 3", trigger=EverySeconds(4))
def task3():
    logger.debug("Task 3 ...")
    time.sleep(4)


if __name__ == '__main__':
    logger.debug("This is the main module.")

    # start worker
    manager1.start()
    manager2.start()

    # create tasks
    Task(name="Task 1", manager=manager2, trigger=AtCreation(delay_for_seconds=10), payload=task1)
    Task(name="Task 2", manager=manager2, trigger=EverySeconds(2), payload=task2)
    Task(name="Task 3", manager=manager2, trigger=EverySeconds(4), payload=task3)

    # enter main loop
    try:
        while True:
            logger.debug("Main loop ...")
            # Manager().loop()
            time.sleep(1)
    except KeyboardInterrupt:
        logger.debug("Keyboard interrupt.")

    manager1.stop()
    manager2.stop()



import contextlib
import logging
from typing import Any

from wiederverwendbar.logger.logger import Logger
from wiederverwendbar.logger.logger_settings import LoggerSettings
from wiederverwendbar.singleton import Singleton

LOGGER_SINGLETON_ORDER = 10


class LoggerSingleton(Logger, metaclass=Singleton, order=LOGGER_SINGLETON_ORDER):
    def __init__(self, name: str,
                 settings: LoggerSettings,
                 use_sub_logger: bool = True,
                 ignored_loggers_equal: list[str] = None,
                 ignored_loggers_like: list[str] = None):
        if ignored_loggers_equal is None:
            ignored_loggers_equal = []
        if ignored_loggers_like is None:
            ignored_loggers_like = []

        super().__init__(name, settings)

        self.ignored_loggers_equal = ignored_loggers_equal
        self.ignored_loggers_like = ignored_loggers_like

        if use_sub_logger:
            logging.setLoggerClass(SubLogger)
            self.configure()

    def is_ignored(self, logger_name: str) -> bool:
        return logger_name in self.ignored_loggers_equal or any([ignored in logger_name for ignored in self.ignored_loggers_like])

    def configure(self):
        for logger in logging.Logger.manager.loggerDict.values():
            if not isinstance(logger, logging.Logger):
                continue
            self.configure_logger(logger)

    def configure_logger(self, logger: logging.Logger):
        if self.is_ignored(logger.name):
            return
        logger.setLevel(self.level)
        logger.parent = self


class SubLogger(logging.Logger):
    def __init__(self, name: str, level=logging.NOTSET):
        self._init = False
        super().__init__(name, level)
        LoggerSingleton().configure_logger(self)
        self._init = True

    def __setattr__(self, key, value):
        if key in ["_init" "init", "_configure_log"]:
            return super().__setattr__(key, value)
        if not self.init:
            return super().__setattr__(key, value)

    @property
    def init(self):
        if hasattr(self, "name"):
            if LoggerSingleton().is_ignored(self.name):
                return False
        return getattr(self, "_init", False)

    @contextlib.contextmanager
    def reconfigure(self):
        if self.init:
            try:
                super().__setattr__("_init", False)
                yield self
            finally:
                super().__setattr__("_init", True)
        else:
            yield self

    def setLevel(self, level):
        if not self.init:
            return super().setLevel(level)

    def addHandler(self, hdlr):
        if not self.init:
            return super().addHandler(hdlr)

    def removeHandler(self, hdlr):
        if not self.init:
            return super().removeHandler(hdlr)

    def addFilter(self, fltr):
        if not self.init:
            return super().addFilter(fltr)

    def removeFilter(self, fltr):
        if not self.init:
            return super().removeFilter(fltr)

    @classmethod
    def get_logger(cls, name: str) -> "SubLogger":
        logger = logging.getLogger(name)
        if not isinstance(logger, SubLogger):
            raise RuntimeError(f"Logger '{name}' is not a {cls.__name__}")
        return logger



import json
import logging
import asyncio
import string
import traceback
import warnings
from typing import Optional, Union

from starlette.requests import Request
from starlette.websockets import WebSocket, WebSocketState
from starlette_admin.exceptions import ActionFailed

logger = logging.getLogger(__name__)


class WebsocketHandler(logging.Handler):
    def __init__(self):
        """
        Create new websocket handler.

        :return: None
        """

        super().__init__()

        self.global_buffer: list[logging.LogRecord] = []  # global_buffer
        self.websockets: dict[WebSocket, list[logging.LogRecord]] = {}  # websocket, websocket_buffer

    def send(self, websocket: WebSocket, record: logging.LogRecord) -> None:
        """
        Send log record to websocket.

        :param websocket: Websocket
        :param record: Log record
        :return: None
        """

        # get extra
        sub_logger_name = getattr(record, "sub_logger")
        command = getattr(record, "command", None)

        command_dict = {"sub_logger": sub_logger_name}

        # check if record is command
        if command is not None:
            command_dict.update(command)
        else:
            msg = self.format(record)
            command_dict.update({"command": "log", "value": msg})

        # convert command to json
        command_json = json.dumps(command_dict)

        # check websocket is connected
        if websocket.client_state != WebSocketState.CONNECTED:
            warnings.warn("Websocket is not connected.")
            return

        # send command message
        try:
            asyncio.run(websocket.send_text(command_json))
        except Exception:
            self.handleError(record)

    def send_all(self) -> None:
        """
        Send all buffered records to all websockets.

        :return: None
        """

        # send buffered records
        for websocket in self.websockets:
            while self.websockets[websocket]:
                buffered_record = self.websockets[websocket].pop(0)
                self.send(websocket, buffered_record)

    def emit(self, record: logging.LogRecord) -> None:
        """
        Emit log record.

        :param record: Log record
        :return: None
        """

        # add record to global buffer
        self.global_buffer.append(record)

        # add record to websocket buffer
        for websocket in self.websockets:
            self.websockets[websocket].append(record)

        # send all
        self.send_all()

    def add_websocket(self, websocket: WebSocket):
        """
        Add websocket to websocket handler. All global buffer will be sent to websocket buffer. After that, all buffered records will be sent to websocket.

        :param websocket: Websocket
        :return: None
        """

        # check if websocket already exists
        if websocket in self.websockets:
            raise ValueError("Websocket already exists.")

        # add websocket to websocket buffer
        self.websockets[websocket] = []

        # push all global buffer to websocket buffer
        for record in self.global_buffer:
            self.websockets[websocket].append(record)

        # send all
        self.send_all()

    def remove_websocket(self, websocket: WebSocket):
        """
        Remove websocket from websocket handler. All buffered records will be sent to websocket.
        :param websocket: Websocket
        :return: None
        """

        # check if websocket exists
        if websocket not in self.websockets:
            raise ValueError("Websocket not exists.")

        # send all
        self.send_all()

        # remove websocket from websocket buffer
        self.websockets.pop(websocket)


class ActionSubLogger(logging.Logger):
    def __init__(self, action_logger: "ActionLogger", name: str, title: Optional[str] = None):
        """
        Create new action sub logger.

        :param action_logger: Action logger
        :param name: Name of sub logger. Only a-z, A-Z, 0-9, - and _ are allowed.
        :param title: Title of sub logger. Visible in frontend.
        """

        super().__init__(name=action_logger.action_log_key + "." + name)

        # validate name
        if not name:
            raise ValueError("Name must not be empty.")
        for char in name:
            if char not in string.ascii_letters + string.digits + "-" + "_":
                raise ValueError("Invalid character in name. Only a-z, A-Z, 0-9, - and _ are allowed.")

        if title is None:
            title = name
        self._title = title
        self._action_logger = action_logger
        self._steps: Optional[int] = None
        self._step: int = 0
        self._websockets: list[WebSocket] = []

        # check if logger already exists
        if self.is_logger_exist(name=self.name):
            raise ValueError("ActionSubLogger already exists.")

        # create websocket handler
        websocket_handler = WebsocketHandler()
        self.addHandler(websocket_handler)

        # add logger to logger manager
        logging.root.manager.loggerDict[self.name] = self

        # start sub logger
        self._command("start", self.title)

    def __del__(self):
        if not self.exited:
            self.exit()

    @classmethod
    def _get_logger(cls, name: str) -> Optional["ActionSubLogger"]:
        """
        Get logger by name.

        :param name: Name of logger.
        :return: Logger
        """

        # get all logger
        all_loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]

        # filter action logger
        for _logger in all_loggers:
            if name != _logger.name:
                continue
            if not isinstance(_logger, ActionSubLogger):
                continue
            return _logger
        return None

    @classmethod
    def is_logger_exist(cls, name: str) -> bool:
        """
        Check if logger exists by name.

        :param name: Name of logger.
        :return: True if exists, otherwise False.
        """

        return cls._get_logger(name=name) is not None

    def _extra(self, extra: Optional[dict] = None) -> dict:
        """
        Add extra to log record and check if any key is already in extra.

        :param extra: Extra of log record.
        :return: New extra.
        """

        _extra = {"sub_logger": self.sub_logger_name}

        if extra is not None:
            # check if any key is already in extra
            for key in extra:
                if key in _extra:
                    raise ValueError(f"Key '{key}' is reserved for ActionSubLogger.")
            _extra.update(extra)
        return _extra

    def _command(self, command: str, value: Union[str, int, float, bool, None]) -> None:
        """
        Send command to websocket.

        :param command: Command
        :param value: Value
        :return: None
        """

        record = self.makeRecord(self.name, logging.NOTSET, "", 0, "", (), None, extra=self._extra({"command": {"command": command, "value": value}}))
        self.handle(record)

    def _log(self, level, msg, args, exc_info=None, extra=None, stack_info=False, stacklevel=1) -> None:
        """
        Log message.

        :param level: Log level
        :param msg: Message
        :param args: Arguments
        :param exc_info: Exception info
        :param extra: Extra added to log record
        :param stack_info: Stack info
        :param stacklevel: Stack level
        :return: None
        """

        super()._log(level=level, msg=msg, args=args, exc_info=exc_info, extra=self._extra(extra), stack_info=stack_info, stacklevel=stacklevel)

    def _step_command(self) -> None:
        """
        Send step command to websocket.

        :return: None
        """

        if self.steps is None:
            return
        calculated_progress = round(self.step / self.steps * 100)

        self._command("step", calculated_progress)

    def add_websocket(self, websocket: WebSocket) -> None:
        """
        Add websocket to sub logger.

        :param websocket: Websocket
        :return: None
        """

        # add websocket to websocket handler
        for handler in self.handlers:
            if not isinstance(handler, WebsocketHandler):
                continue
            handler.add_websocket(websocket)

        # add websocket to sub logger
        if websocket in self._websockets:
            return
        self._websockets.append(websocket)

    def remove_websocket(self, websocket: WebSocket):
        """
        Remove websocket from sub logger.

        :param websocket: Websocket
        :return: None
        """

        # remove websocket from sub logger
        for handler in self.handlers:
            if not isinstance(handler, WebsocketHandler):
                continue
            handler.remove_websocket(websocket)

        # remove websocket from sub logger
        if websocket not in self._websockets:
            return
        self._websockets.remove(websocket)

    @property
    def sub_logger_name(self) -> str:
        """
        Get sub logger name.

        :return: Sub logger name.
        """

        return self.name.replace(self._action_logger.action_log_key + ".", "")

    @property
    def title(self) -> str:
        """
        Get title of sub logger.

        :return: Title of sub logger.
        """

        return self._title

    @property
    def steps(self) -> int:
        """
        Get steps of sub logger.

        :return: Steps of sub logger.
        """

        return self._steps

    @steps.setter
    def steps(self, value: int) -> None:
        """
        Set steps of sub logger. Also send step command to websocket.

        :param value: Steps
        :return: None
        """

        if value < 0:
            raise ValueError("Steps must be greater than 0.")
        self._steps = value
        self._step_command()

    @property
    def step(self) -> int:
        """
        Get step of sub logger.

        :return: Step of sub logger.
        """
        return self._step

    @step.setter
    def step(self, value: int) -> None:
        """
        Set step of sub logger. Also send step command to websocket.

        :param value: Step
        :return: None
        """

        self._step = value
        self._step_command()

    def next_step(self) -> None:
        """
        Increase step by 1.

        :return: None
        """

        if self.step >= self.steps:
            return
        self.step += 1

    def finalize(self, success: bool = True, log_level: int = logging.INFO, msg: Optional[str] = None) -> None:
        """
        Finalize sub logger. Also send finalize command to websocket.

        :param success: If True, frontend will show success message. If False, frontend will show error message.
        :param log_level: Log level of finalize message.
        :param msg: Message of finalize message.
        :return: None
        """

        if self.exited:
            raise ValueError("ActionSubLogger already exited.")

        if success:
            if self.steps is not None:
                if self.step < self.steps:
                    self.step = self.steps
        if msg is not None:
            self.log(log_level, msg)

        self._command("finalize", success)
        self.exit()

    def exit(self) -> None:
        """
        Exit sub logger. Also remove websocket from sub logger.

        :return: None
        """

        if self.exited:
            raise ValueError("ActionSubLogger already exited.")

        # remove websockets
        for websocket in self._websockets:
            self.remove_websocket(websocket)

        # remove handler
        for handler in self.handlers:
            self.removeHandler(handler)

        # remove logger from logger manager
        logging.root.manager.loggerDict.pop(self.name, None)

    @property
    def exited(self) -> bool:
        """
        Check if sub logger is exited.

        :return: True if exited, otherwise False.
        """

        return not self.is_logger_exist(name=self.name)


class ActionSubLoggerContext:
    def __init__(self,
                 action_logger: "ActionLogger",
                 name: str,
                 title: Optional[str] = None,
                 log_level: Optional[int] = None,
                 parent: Optional[logging.Logger] = None,
                 formatter: Optional[logging.Formatter] = None,
                 steps: Optional[int] = None,
                 finalize_on_success_log_level: int = logging.INFO,
                 finalize_on_success_msg: Optional[str] = None,
                 show_errors: Optional[bool] = None):
        """
        Create new action sub logger context manager.

        :param action_logger: Action logger
        :param name: Name of sub logger. Only a-z, A-Z, 0-9, - and _ are allowed.
        :param title: Title of sub logger. Visible in frontend.
        :param log_level: Log level of sub logger. If None, parent log level will be used. If parent is None, action logger log level will be used.
        :param parent: Parent logger. If None, action logger parent will be used.
        :param formatter: Formatter of sub logger. If None, action logger formatter will be used.
        :param steps: Steps of sub logger.
        :param finalize_on_success_log_level: Log level of finalize message if success.
        :param finalize_on_success_msg: Message of finalize message if success.
        :param show_errors: Show errors in frontend. If None, action logger show_errors will be used.
        """

        # create sub logger
        self.sub_logger = action_logger.new_sub_logger(name=name, title=title, log_level=log_level, parent=parent, formatter=formatter, steps=steps)

        self.finalize_on_success_log_level = finalize_on_success_log_level
        self.finalize_on_success_msg = finalize_on_success_msg
        if show_errors is None:
            show_errors = action_logger.show_errors
        self.show_errors = show_errors

    def __enter__(self) -> "ActionSubLogger":
        return self.sub_logger

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.sub_logger.exited:
            if exc_type is None:
                self.sub_logger.finalize(success=True, log_level=self.finalize_on_success_log_level, msg=self.finalize_on_success_msg)
            else:
                if self.show_errors:
                    # get exception string
                    tb_str = traceback.format_exc()
                    self.sub_logger.finalize(success=False, log_level=logging.ERROR, msg=tb_str)
                else:
                    self.sub_logger.finalize(success=False, log_level=logging.ERROR, msg="Something went wrong.")


class ActionLogger:
    _action_loggers: list["ActionLogger"] = []

    def __init__(self,
                 action_log_key_request_or_websocket: Union[str, Request],
                 log_level: Optional[int] = None,
                 parent: Optional[logging.Logger] = None,
                 formatter: Optional[logging.Formatter] = None,
                 show_errors: bool = True,
                 wait_for_websocket: bool = True,
                 wait_for_websocket_timeout: int = 5):
        """
        Create new action logger.

        :param action_log_key_request_or_websocket: Action log key, request or websocket.
        :param log_level: Log level of action logger. If None, parent log level will be used. If parent is None, logging.INFO will be used.
        :param parent: Parent logger. If None, logger will be added to module logger.
        :param formatter: Formatter of action logger. If None, default formatter will be used.
        """

        self.action_log_key = self.get_action_key(action_log_key_request_or_websocket)
        self.show_errors = show_errors

        # get parent logger
        if parent is None:
            parent = logger
        self.parent = parent

        # set log level
        if log_level is None:
            if parent is None:
                log_level = logging.INFO
            else:
                log_level = parent.level
        self.log_level = log_level

        # set formatter
        if formatter is None:
            formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s", "%Y-%m-%d - %H:%M:%S")
        self.formatter = formatter

        self._websockets: list[WebSocket] = []
        self._sub_logger: list[ActionSubLogger] = []

        # add action logger to action loggers
        self._action_loggers.append(self)

        # wait for websocket
        if wait_for_websocket:
            current_try = 0
            while len(self._websockets) == 0:
                if current_try >= wait_for_websocket_timeout:
                    raise ValueError("No websocket connected.")
                current_try += 1
                logger.debug(f"[{current_try}/{wait_for_websocket_timeout}] Waiting for websocket...")
                asyncio.run(asyncio.sleep(1))

    def __enter__(self) -> "ActionLogger":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.exited:
            self.exit()

        # get exception string
        if exc_type is not None and self.show_errors:
            te = traceback.TracebackException(type(exc_type), exc_val, exc_tb)
            efs = te.stack[-1]
            exception_str = f"{exc_type.__name__}: {exc_val}"
            # add line number
            if exc_tb is not None:
                exception_str += f" at line {efs.lineno} in {efs.filename}"

            # raise ActionFailed
            raise ActionFailed(exception_str)

    def __del__(self):
        if not self.exited:
            self.exit()

    @classmethod
    async def get_logger(cls, action_log_key_request_or_websocket: Union[str, Request, WebSocket]) -> Optional["ActionLogger"]:
        """
        Get action logger by action log key or request.

        :param action_log_key_request_or_websocket: Action log key, request or websocket.
        :return: Action logger.
        """

        for _action_logger in cls._action_loggers:
            if _action_logger.action_log_key == cls.get_action_key(action_log_key_request_or_websocket):
                return _action_logger
        return None

    @classmethod
    def get_action_key(cls, action_log_key_request_or_websocket: Union[str, Request, WebSocket]) -> str:
        """
        Get action log key from request or websocket.

        :param action_log_key_request_or_websocket: Action log key, request or websocket.
        :return: Action log key.
        """

        if isinstance(action_log_key_request_or_websocket, Request):
            action_log_key = action_log_key_request_or_websocket.query_params.get("actionLogKey", None)
            if action_log_key is None:
                raise ValueError("No action log key provided.")
        elif isinstance(action_log_key_request_or_websocket, WebSocket):
            action_log_key = action_log_key_request_or_websocket.path_params.get("action_log_key", None)
            if action_log_key is None:
                raise ValueError("No action log key provided.")
        elif isinstance(action_log_key_request_or_websocket, str):
            action_log_key = action_log_key_request_or_websocket
        else:
            raise ValueError("Invalid action log key or request.")
        return action_log_key

    @classmethod
    async def wait_for_logger(cls, action_log_key_request_or_websocket: Union[str, Request, WebSocket], timeout: int = 5) -> "ActionLogger":
        """
        Wait for action logger to be created by WebSocket connection. If action logger not found, a dummy logger will be created or an error will be raised.

        :param action_log_key_request_or_websocket: Action log key, request or websocket.
        :param timeout: Timeout in seconds.
        :return: Action logger.
        """

        # get action logger
        action_logger = None
        current_try = 0
        while current_try < timeout:
            action_logger = await cls.get_logger(cls.get_action_key(action_log_key_request_or_websocket))

            # if action logger found, break
            if action_logger is not None:
                break
            current_try += 1
            logger.debug(f"[{current_try}/{timeout}] Waiting for action logger...")
            await asyncio.sleep(1)

        # check if action logger finally found
        if action_logger is None:
            raise ValueError("ActionLogger not found.")

        return action_logger

    def add_websocket(self, websocket: WebSocket) -> None:
        """
        Add websocket to action logger.

        :param websocket: Websocket
        :return: None
        """

        # add websocket to sub loggers
        for sub_logger in self._sub_logger:
            sub_logger.add_websocket(websocket)

        # add websocket to action logger
        if websocket in self._websockets:
            return
        self._websockets.append(websocket)

    def remove_websocket(self, websocket: WebSocket) -> None:
        """
        Remove websocket from action logger.

        :param websocket: Websocket
        :return: None
        """

        # remove websocket from sub loggers
        for sub_logger in self._sub_logger:
            sub_logger.remove_websocket(websocket)

        # remove websocket from action logger
        if websocket not in self._websockets:
            return
        self._websockets.remove(websocket)

    def new_sub_logger(self,
                       name: str,
                       title: Optional[str] = None,
                       log_level: Optional[int] = None,
                       parent: Optional[logging.Logger] = None,
                       formatter: Optional[logging.Formatter] = None,
                       steps: Optional[int] = None) -> ActionSubLogger:
        """
        Create new sub logger.

        :param name: Name of sub logger. Only a-z, A-Z, 0-9, - and _ are allowed.
        :param title: Title of sub logger. Visible in frontend.
        :param log_level: Log level of sub logger. If None, parent log level will be used. If parent is None, action logger log level will be used.
        :param parent: Parent logger. If None, action logger parent will be used.
        :param formatter: Formatter of sub logger. If None, action logger formatter will be used.
        :param steps: Steps of sub logger. If None, no steps will be shown.
        :return:
        """

        try:
            self.get_sub_logger(sub_logger_name=name)
        except ValueError:
            pass

        # create sub logger
        sub_logger = ActionSubLogger(action_logger=self, name=name, title=title)

        # set parent logger
        if parent is None:
            parent = self.parent
        sub_logger.parent = parent

        # set log level
        if log_level is None:
            if parent is None:
                log_level = self.log_level
            else:
                log_level = parent.level
        sub_logger.setLevel(log_level)

        # set formatter
        if formatter is None:
            formatter = self.formatter
        for handler in sub_logger.handlers:
            handler.setFormatter(formatter)

        # set steps
        if steps is not None:
            sub_logger.steps = steps

        # add websocket to sub logger
        for websocket in self._websockets:
            sub_logger.add_websocket(websocket)

        self._sub_logger.append(sub_logger)
        return sub_logger

    def get_sub_logger(self, sub_logger_name: str) -> ActionSubLogger:
        """
        Get sub logger by name.

        :param sub_logger_name: Name of sub logger.
        :return:
        """

        if self.exited:
            raise ValueError("ActionLogger already exited.")

        # check if sub logger already exists
        for sub_logger in self._sub_logger:
            if sub_logger.sub_logger_name == sub_logger_name:
                return sub_logger
        raise ValueError("Sub logger not found.")

    def sub_logger(self,
                   name: str,
                   title: Optional[str] = None,
                   log_level: Optional[int] = None,
                   parent: Optional[logging.Logger] = None,
                   formatter: Optional[logging.Formatter] = None,
                   steps: Optional[int] = None,
                   finalize_on_success_log_level: int = logging.INFO,
                   finalize_on_success_msg: Optional[str] = None,
                   show_errors: Optional[bool] = None) -> ActionSubLoggerContext:
        """
        Sub logger context manager.

        :param name: Name of sub logger. Only a-z, A-Z, 0-9, - and _ are allowed.
        :param title: Title of sub logger. Visible in frontend.
        :param log_level: Log level of sub logger. If None, parent log level will be used. If parent is None, action logger log level will be used.
        :param parent: Parent logger. If None, action logger parent will be used.
        :param formatter: Formatter of sub logger. If None, action logger formatter will be used.
        :param steps: Steps of sub logger.
        :param finalize_on_success_log_level: Log level of finalize message if success.
        :param finalize_on_success_msg: Message of finalize message if success.
        :param show_errors: Show errors in frontend. If None, action logger show_errors will be used.
        :return:
        """

        return ActionSubLoggerContext(action_logger=self,
                                      name=name,
                                      title=title,
                                      log_level=log_level,
                                      parent=parent,
                                      formatter=formatter,
                                      steps=steps,
                                      finalize_on_success_log_level=finalize_on_success_log_level,
                                      finalize_on_success_msg=finalize_on_success_msg,
                                      show_errors=show_errors)

    def exit(self):
        """
        Exit action logger. Also remove all websockets and sub loggers.

        :return: None
        """

        if self.exited:
            raise ValueError("ActionLogger already exited.")

        # remove websockets
        for websocket in self._websockets:
            self.remove_websocket(websocket)

        # exit sub loggers
        for sub_logger in self._sub_logger:
            if not sub_logger.exited:
                sub_logger.exit()

        # remove action logger from action loggers
        self._action_loggers.remove(self)

    @property
    def exited(self) -> bool:
        """
        Check if action logger is exited.

        :return: True if exited, otherwise False.
        """

        return self not in self._action_loggers



from typing import Any, Dict

from starlette.requests import Request
from starlette_admin import action
from starlette_admin.contrib.mongoengine import ModelView as BaseModelView
from starlette_admin import RequestAction


class FixedModelView(BaseModelView):
    async def serialize(
            self,
            obj: Any,
            request: Request,
            action: RequestAction,
            include_relationships: bool = True,
            include_select2: bool = False,
    ) -> Dict[str, Any]:
        result = await super().serialize(obj, request, action, include_relationships, include_select2)
        result["id"] = str(result["id"])
        return result



import encodings
import logging
from enum import Enum
from pathlib import Path
from typing import Any, Optional, Union

from pydantic import BaseModel, Field, field_validator


class LoggerSettings(BaseModel):
    class LogLevels(str, Enum):
        """
        Log levels
        """

        CRITICAL = "CRITICAL"
        FATAL = "FATAL"
        ERROR = "ERROR"
        WARNING = "WARNING"
        INFO = "INFO"
        DEBUG = "DEBUG"

    log_level: LogLevels = Field(default=LogLevels.WARNING,
                                 title="Log Level",
                                 description="The log level")

    log_console: bool = Field(default=True,
                              title="Console Logging",
                              description="Whether to log to the console")
    log_console_level: Optional[LogLevels] = Field(default=None,
                                                   title="Console Log Level",
                                                   description="The log level for the console")
    log_console_format: str = Field(default="%(name)s - %(message)s",
                                    title="Console Log Format",
                                    description="The log format for the console")
    log_console_width: int = Field(default=80,
                                   title="Console Width",
                                   ge=0,
                                   description="The width of the console")

    class TerminalOutFile(str, Enum):
        """
        Terminal output file
        """

        STDOUT = "stdout"
        STDERR = "stderr"

    log_console_outfile: TerminalOutFile = Field(default=TerminalOutFile.STDOUT,
                                                 title="Console Outfile",
                                                 description="The console outfile")
    log_console_rich_markup: bool = Field(default=True,
                                          title="Rich Markup",
                                          description="Whether to use rich markup in the console")
    log_console_rich_show_time: bool = Field(default=False,
                                             title="Show Time in Console",
                                             description="Whether to show the time in the console")
    log_console_rich_show_level: bool = Field(default=True,
                                              title="Show Level in Console",
                                              description="Whether to show the level in the console")
    log_console_rich_show_path: bool = Field(default=True,
                                             title="Show Path in Console",
                                             description="Whether to show the path in the console")
    log_file: bool = Field(default=False,
                           title="File Logging",
                           description="Whether to log to a file")
    log_file_path: Optional[Path] = Field(default=None,
                                          title="Log File Path",
                                          description="The path of the log file")
    log_file_level: Optional[LogLevels] = Field(default=None,
                                                title="File Log Level",
                                                description="The log level for the file")
    log_file_format: str = Field(default="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
                                 title="File Log Format",
                                 description="The log format for the file")

    class FileMode(str, Enum):
        """
        File mode
        """

        a = "a"
        w = "w"

    log_file_mode: FileMode = Field(default=FileMode.a,
                                    title="File Mode",
                                    description="The file mode")
    log_file_max_bytes: int = Field(default=1024 * 1024 * 10,
                                    title="Max File Size",
                                    ge=1024,
                                    description="The maximum size of the log file. Default is 10MB")
    log_file_backup_count: int = Field(default=5,
                                       title="Backup Log Files",
                                       description="The number of backup log files to keep")
    log_file_encoding: str = Field(default="utf-8",
                                   title="File Encoding",
                                   description="The encoding of the log file")

    log_file_delay: bool = Field(default=False,
                                 title="Delay File Logging",
                                 description="Whether to delay the file logging")
    log_file_archive_backup_count: int = Field(default=5,
                                               title="Backup Log Archives",
                                               ge=0,
                                               description="The number of backup log archives to keep")

    def __init__(self, /, **data: Any):
        super().__init__(**data)

        if self.log_console_level is None:
            self.log_console_level = self.log_level

        if self.log_file_level is None:
            self.log_file_level = self.log_level

    @field_validator("log_level", "log_console_level", "log_file_level", mode="before")
    def validate_log_level(cls, value: Union[int, str]) -> str:
        if isinstance(value, int):
            value = logging.getLevelName(value)
        return value

    @field_validator("log_file_encoding")
    def validate_log_file_encoding(cls, value):
        # check if encoding is available
        available_encodings = [encoding_name.replace("_", "-") for encoding_name in encodings.aliases.aliases.values()]
        if value not in available_encodings:
            raise ValueError(f"Encoding '{value}' is not available. Available encodings: {', '.join(available_encodings)}")
        return value



from jinja2 import PackageLoader

from wiederverwendbar.starlette_admin import MultiPathAdmin


class TestAdmin1(MultiPathAdmin):
    static_files_packages = [("wiederverwendbar", "starlette_admin/statics")]
    template_packages = [PackageLoader("wiederverwendbar", "starlette_admin/mongoengine/generic_embedded_document_field")]


class TestAdmin2(TestAdmin1):
    static_files_packages = [("wiederverwendbar", "starlette_admin/statics/js")]
    template_packages = [PackageLoader("wiederverwendbar", "starlette_admin/mongoengine/generic_embedded_document_field/templates")]


base_admin = MultiPathAdmin(title="Test Admin")
admin1 = TestAdmin1(title="Test Admin")
admin2 = TestAdmin2(title="Test Admin")



import logging
import asyncio

import uvicorn
from starlette.applications import Starlette
from starlette.responses import HTMLResponse
from starlette.routing import Route
from starlette.requests import Request
from starlette_admin.contrib.mongoengine import Admin, ModelView
from starlette_admin.actions import action
from mongoengine import connect, Document, StringField

from wiederverwendbar.starlette_admin import ActionLogAdmin, ActionLogger

logger = logging.getLogger(__name__)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
logger.addHandler(ch)
logger.setLevel(logging.DEBUG)

# connect to database
connect("test",
        host="localhost",
        port=27017)

# Create starlette app
app = Starlette(
    routes=[
        Route(
            "/",
            lambda r: HTMLResponse("<a href=/admin/>Click me to get to Admin!</a>"),
        ),
    ],
)


class MyAdmin(Admin, ActionLogAdmin):
    ...


# Create admin
admin = MyAdmin(title="Test Admin")


class Test(Document):
    meta = {"collection": "test"}

    test_str = StringField()


class TestView(ModelView):
    def __init__(self):
        super().__init__(document=Test, icon="fa fa-server", name="Test", label="Test")

    actions = ["delete", "test_action_normal", "test_action_action_log"]

    @action(name="test_action_normal",
            text="Test Action - Normal")
    # confirmation="Möchtest du die Test Aktion durchführen?",
    # icon_class="fa-regular fa-network-wired",
    # submit_btn_text="Ja, fortsetzen",
    # submit_btn_class="btn-success")
    async def test_action_normal(self, request: Request, pk: list[str]) -> str:
        await asyncio.sleep(2)

        return "Test Aktion erfolgreich."

    @action(name="test_action_action_log",
            text="Test Action - Action Log")
    # confirmation="Möchtest du die Test Aktion durchführen?",
    # icon_class="fa-regular fa-network-wired",
    # submit_btn_text="Ja, fortsetzen",
    # submit_btn_class="btn-success")
    async def test_action_action_log(self, request: Request, pk: list[str]) -> str:
        with ActionLogger(request) as action_logger:
            # use context manager to ensure that the logger is finalized
            with action_logger.sub_logger("sub_action_1", "Sub Action 1", steps=3) as sub_logger:
                sub_logger.info("Test Aktion startet ...")
                sub_logger.debug("Debug")
                sub_logger.info("Test Aktion step 1")
                await asyncio.sleep(2)
                sub_logger.next_step()
                sub_logger.info("Test Aktion step 2")
                _ = 1 / 0  # raise exception
                await asyncio.sleep(2)
                sub_logger.next_step()
                sub_logger.steps += 100
                for i in range(1, 100):
                    sub_logger.info(f"Test Aktion step 2 - {i}")
                    sub_logger.next_step()
                    await asyncio.sleep(0.1)
                sub_logger.info("Test Aktion step 3")
                await asyncio.sleep(2)

            sub_action_2_logger = action_logger.new_sub_logger("sub_action_2", "Sub Action 2", steps=3)
            sub_action_3_logger = action_logger.new_sub_logger("sub_action_3", "Sub Action 3", steps=3)
            sub_action_2_logger.info("Test Aktion startet ...")
            sub_action_3_logger.info("Test Aktion startet ...")
            await asyncio.sleep(2)
            sub_action_2_logger.next_step()
            sub_action_3_logger.next_step()
            await asyncio.sleep(2)
            sub_action_2_logger.next_step()
            sub_action_3_logger.next_step()
            sub_action_2_logger.finalize(success=False, msg="Test Aktion fehlgeschlagen.")
            sub_action_3_logger.finalize(success=True, msg="Test Aktion erfolgreich.")

        return "Test Aktion erfolgreich."


# Add views to admin#
admin.add_view(TestView())

# Mount admin to app
admin.mount_to(app)

if __name__ == "__main__":
    uvicorn.run(app, host="0.0.0.0", port=8000)



from jinja2 import PackageLoader
from starlette_admin.contrib.mongoengine import Admin as MongoengineAdmin

from wiederverwendbar.starlette_admin.admin import MultiPathAdmin, FormMaxFieldsAdmin


class GenericEmbeddedAdmin(MongoengineAdmin, FormMaxFieldsAdmin, MultiPathAdmin):
    form_max_fields = 10000
    static_files_packages = [("wiederverwendbar", "starlette_admin/mongoengine/generic_embedded_document_field/statics")]
    template_packages = [PackageLoader("wiederverwendbar", "starlette_admin/mongoengine/generic_embedded_document_field/templates")]



import os

import hashlib

from wiederverwendbar.pydantic.security.hashed_password_model import HashedPasswordModel


def hash_password(password: str,
                  encoding: str = "utf-8",
                  hash_function: str = "sha256",
                  interactions: int = 100000,
                  key_length: int = 128,
                  salt: str = None) -> HashedPasswordModel:
    if salt is None:
        salt = os.urandom(32)
        salt = salt.hex()
    salt_encoded = salt.encode(encoding)
    password_encoded = password.encode(encoding)
    hashed_password_encoded = hashlib.pbkdf2_hmac(hash_function, password_encoded, salt_encoded, interactions, key_length)  # generate hash
    hashed_password = hashed_password_encoded.hex()
    out = HashedPasswordModel(
        encoding=encoding,
        hash_function=hash_function,
        interactions=interactions,
        key_length=key_length,
        salt=salt,
        hashed_password=hashed_password
    )
    return out
